#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <GL/glew.h>	/* glew.h prin apo opoiodipote GL i glut header file */
#include <GL/glut.h>
#include "mesh.h"
#include "sdr.h"
#include "image_rgba.h"

int init();
void display();
void draw_floor();
void draw_teapot();
void idle();
void reshape(int x, int y);
void keyboard(unsigned char key, int x, int y);
void mouse(int button, int state, int x, int y);
void motion(int x, int y);
unsigned int load_texture_rgba(const char *fname);

/* --- global variables --- */
float cam_theta, cam_phi = 25, cam_dist = 8;

struct mesh *dragon;
unsigned int sdrprog;


int main(int argc, char **argv)
{
	glutInit(&argc, argv);
	glutInitWindowSize(800, 600);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("OpenGL askisi");

	glutDisplayFunc(display);
	glutIdleFunc(idle);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	if(init() == -1) {
		return 1;
	}

	glutMainLoop();
	return 0;
}


/* --- initialization, ekteleite stin arxi apo tin main, epistrefei -1 an yparksei provlima --- */
int init()
{
	/* panta otan xrisimopoioume GLEW, prota kaloume tin glewInit */
	glewInit();

	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	dragon = load_mesh("dragon.obj");
	if(!dragon) {
		return -1;
	}

	sdrprog = create_program_load("vertex.glsl", "pixel.glsl");
	if(!sdrprog) {
		return -1;
	}

	return 0;
}

void display()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	/* --- setup view (camera) matrix --- */
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0, 0, -cam_dist);
	glRotatef(cam_phi, 1, 0, 0);
	glRotatef(cam_theta, 0, 1, 0);
	/* --- end of view matrix ---*/

	/* --- set light position --- */
	float lpos[] = {-10, 20, 10, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);

	/* --- render all objects, thetontas prota to world matrix tous --- */

	// drakos
	// TODO 3: o shader tou kokkinou drakou na ypologizei fotismo diffuse (bonus: kai specular)
	glUseProgram(sdrprog);
	render_mesh(dragon);
	glUseProgram(0);

	// tsagieres
	// TODO 1: render 4 tsagieres stis gonies tou patomatos (use draw_teapot)

	// patoma
	// TODO 2: to patoma na exei texture (floor.tga) tiled 4 fores
	draw_floor();

	/* --- sto telos glutSwapBuffers gia na emfanistei to apotelesma --- */
	glutSwapBuffers();
	assert(glGetError() == GL_NO_ERROR);
}

void draw_floor()
{
	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glVertex3f(-8, 0, 8);
	glVertex3f(8, 0, 8);
	glVertex3f(8, 0, -8);
	glVertex3f(-8, 0, -8);
	glEnd();
}

void draw_teapot()
{
	glFrontFace(GL_CW);
	glutSolidTeapot(1.0);
	glFrontFace(GL_CCW);
}

void idle()
{
	glutPostRedisplay();
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	/* --- set the projection matrix --- */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 0.5, 500.0);
}

void keyboard(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
	case 'q':
		exit(0);
	}
}


int bnstate[16];
int prev_x, prev_y;

void mouse(int button, int state, int x, int y)
{
	/* kratame tin katastasi ton pliktron sto array bnstate */
	int bidx = button - GLUT_LEFT_BUTTON;
	if(state == GLUT_DOWN) {
		bnstate[bidx] = 1;

		/* thetoume to prev_x kai prev_y sto simio tou pontikiou otan patithike kapoio pliktro */
		prev_x = x;
		prev_y = y;
	} else {
		bnstate[bidx] = 0;
	}
}

void motion(int x, int y)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	/* an patithike to aristero koumpi */
	if(bnstate[0]) {
		cam_theta += (float)dx * 0.5;
		cam_phi += (float)dy * 0.5;

		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;
	}

	/* an patithike to deksi koumpi */
	if(bnstate[2]) {
		cam_dist += (float)dy * 0.1;
		if(cam_dist < 0) cam_dist = 0;
	}
}

unsigned int load_texture_rgba(const char *fname)
{
	struct image_rgba *img = load_image_rgba(fname);
	if(!img) {
		return 0;
	}


	unsigned int tex;
	glGenTextures(1, &tex);
	glBindTexture(GL_TEXTURE_2D, tex);

	// enable mip-mapping if mipmap generation is available
	if(GLEW_SGIS_generate_mipmap) {
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, 1);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	} else {
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, img->width, img->height, 0, GL_RGBA, GL_UNSIGNED_BYTE, img->pixels);
	free_image_rgba(img);

	return tex;
}
